package dev.pfaff.unfettered;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

import static dev.pfaff.unfettered.Log2.log2nlz;
import static dev.pfaff.unfettered.Util.reflectUnchecked;
import static java.lang.invoke.MethodType.methodType;

/**
 * <h2>Breaking free of the chains:</h2>
 *
 * <p>This class goes through a lengthy process to shatter the restraints of JPMS without requiring the end-user,
 * launcher, or mod loader to inject any JVM args.</p>
 *
 * <p>First, we get a reference to {@link sun.misc.Unsafe#theUnsafe} via reflection, which is not yet locked down by
 * the ongoing encapsulation crusade.</p>
 *
 * <p>Then, using this reference, we get the static field base and offset of {@link MethodHandles.Lookup#IMPL_LOOKUP}.
 * As far as I am aware, this is the only instance of {@link MethodHandles.Lookup} with an
 * {@link MethodHandles.Lookup#allowedModes allowedModes} value of {@link MethodHandles.Lookup#TRUSTED TRUSTED}. We then
 * check the OOP pointer width and, equipped with these, use our reference to {@link sun.misc.Unsafe#theUnsafe} to copy
 * the OOP pointer stored in the static field into a 1-element array on the LVT.</p>
 *
 * <p>Now might be a good time to interject and point out how pointless it is for the JDK to hide
 * {@link MethodHandles.Lookup#allowedModes} (see {@code MethodHandles.java} line 1433:
 * {@code Reflection.registerFieldsToFilter(Lookup.class, Set.of("lookupClass", "allowedModes"));}) from reflection
 * without also hiding {@link MethodHandles.Lookup#IMPL_LOOKUP}.</p>
 *
 * <p>Now equipped with an unrestricted {@link MethodHandles.Lookup}, we get a {@link MethodHandle} for
 * {@link Module#implAddExportsOrOpens(String, Module, boolean, boolean)}.</p>
 *
 * <p>Phew, almost there! Now we get a reference to this and the {@code java.base} {@link Module}s via
 * {@link Class#getModule()} ({@code UnsafeUtil.class.getModule()} and {@code String.class.getModule()}). Next, we use
 * that handle we just got to export the {@code jdk.internal.misc} package in the {@code java.base} module to this
 * module.</p>
 *
 * <p>Finally, references to {@code jdk.internal.misc.Unsafe#theUnsafe} and a usable {@link MethodHandles.Lookup} (the
 * one we borrowed earlier doesn't have access to any of <em>our</em> classes): For the former, we simply call
 * {@code jdk.internal.misc.Unsafe#getUnsafe()}. It's a bit more complicated for the latter, but nothing compared to
 * what we've accomplished thus far.</p>
 *
 * <p>First, we set a {@code static final} field to the value of {@link MethodHandles#lookup()}. Then, we get the
 * offset of the {@link MethodHandles.Lookup#allowedModes allowedModes} field using our shiny new reference to
 * {@code jdk.internal.misc.Unsafe#theUnsafe} and then, once again using {@code jdk.internal.misc.Unsafe#theUnsafe}, set
 * the value to {@link MethodHandles.Lookup#TRUSTED}.</p>
 *
 * <p>And that's it! We now have full access to the classpath and the memory space of the JVM.</p>
 */
public final class Unfettered {
	// needed for bootstrap
	private static final sun.misc.Unsafe theSunUnsafe = reflectUnchecked(() -> {
		var f = sun.misc.Unsafe.class.getDeclaredField("theUnsafe");
		f.setAccessible(true);
		return (sun.misc.Unsafe) f.get(null);
	});

	public static final MethodHandles.Lookup theJdkTrustedLookup;

	// see the class documentation for a detailed breakdown of this initialization.
	static {
		// we cannot simply call f.setAccessible(true) because the package is not opened to us.
		final Field f = reflectUnchecked(() -> MethodHandles.Lookup.class.getDeclaredField("IMPL_LOOKUP"));
		var IMPL_LOOKUP_BASE = theSunUnsafe.staticFieldBase(f);
		var IMPL_LOOKUP_OFFSET = theSunUnsafe.staticFieldOffset(f);

		final MethodHandles.Lookup[] borrowedLookup = new MethodHandles.Lookup[1];
		if (sun.misc.Unsafe.ARRAY_OBJECT_INDEX_SCALE == Integer.BYTES) {
			theSunUnsafe.putInt(borrowedLookup,
					sun.misc.Unsafe.ARRAY_OBJECT_BASE_OFFSET,
					theSunUnsafe.getInt(IMPL_LOOKUP_BASE, IMPL_LOOKUP_OFFSET));
		} else {
			assert sun.misc.Unsafe.ARRAY_OBJECT_INDEX_SCALE == Long.BYTES
					: "OOPs are neither 32-bit nor 64-bit: pointerWidth=" + sun.misc.Unsafe.ADDRESS_SIZE + ", " +
					"arrayObjectIndexScale=" + sun.misc.Unsafe.ARRAY_OBJECT_INDEX_SCALE;
			theSunUnsafe.putLong(borrowedLookup,
					sun.misc.Unsafe.ARRAY_OBJECT_BASE_OFFSET,
					theSunUnsafe.getLong(IMPL_LOOKUP_BASE, IMPL_LOOKUP_OFFSET));
		}

		theJdkTrustedLookup = borrowedLookup[0];
	}

	public static final Class<?> C_Unsafe = reflectUnchecked(() -> theJdkTrustedLookup.findClass("jdk.internal.misc.Unsafe"));
	public static final Object theUnsafe;

	static {
		try {
			theUnsafe = theJdkTrustedLookup.findStatic(C_Unsafe, "getUnsafe", methodType(C_Unsafe))
					.asType(methodType(Object.class))
					.invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private static final long LOOKUP_ALLOWED_MODES_OFFSET = UnsafeWrapper.objectFieldOffset(MethodHandles.Lookup.class, "allowedModes");

	private static final MethodHandle MH_Lookup_constructor = reflectUnchecked(() -> {
		return theJdkTrustedLookup.findConstructor(MethodHandles.Lookup.class, methodType(void.class, Class.class));
	});

	/**
	 * Produces a new {@link MethodHandles.Lookup} with {@link MethodHandles.Lookup#TRUSTED} access.
	 */
	private static MethodHandles.Lookup makeTrustedLookup(Class<?> lookupClass) {
		MethodHandles.Lookup l;
		try {
			l = (MethodHandles.Lookup) MH_Lookup_constructor.invokeExact(lookupClass);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		UnsafeWrapper.putInt(l, LOOKUP_ALLOWED_MODES_OFFSET, -1);
		return l;
	}

	/**
	 * Method handle lookup with the {@code allowedModes} field set to {@link MethodHandles.Lookup#TRUSTED}, giving it
	 * unrestricted access to every class in the JVM.
	 * <p>
	 * <strong>Note that this will not be trusted until part of the way through static init</strong>
	 */
	public static final MethodHandles.Lookup theTrustedLookup = makeTrustedLookup(Unfettered.class);

	public static MethodHandles.Lookup getTrustedLookup() {
		return theTrustedLookup;
	}

	/**
	 * Forces this class to be static-initialized. Calling this method more than once will have no effect. Calling this
	 * method is not even necessary before using any of the exposed unsafe APIs.
	 */
	public static void forceInit() {
	}

	private static final MethodHandle MH_implAddExportsOrOpens = reflectUnchecked(() -> theJdkTrustedLookup.findSpecial(
			Module.class,
			"implAddExportsOrOpens",
			methodType(
					Void.TYPE,
					String.class,
					Module.class,
					Boolean.TYPE,
					Boolean.TYPE
			),
			Module.class
	));

	/**
	 * Exports the {@code pkgName} from the {@code module} to {@code target}. If {@code opens} is {@code true}, then the
	 * {@code target} will have access to package-private members in addition to public ones.
	 */
	public static void export(Module module, String pkgName, boolean opens, Module target) {
		try {
			MH_implAddExportsOrOpens.invokeExact(module, pkgName, target, opens, true);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * A handle to the private, no security checks method for getting the fields. DO NOT modify the returned array.
	 */
	private static final MethodHandle MH_getDeclaredFieldsQuick = reflectUnchecked(() -> {
		return getTrustedLookup().findVirtual(Class.class,
				"privateGetDeclaredFields",
				methodType(Field[].class, boolean.class));
	});

	private static Field[] getPublicFieldsQuick(Class<?> clazz) {
		try {
			return (Field[]) MH_getDeclaredFieldsQuick.invokeExact(clazz, true);
		} catch (Throwable e) {
			// should never happen
			throw new RuntimeException(e);
		}
	}

	private static Field[] getDeclaredFieldsQuick(Class<?> clazz) {
		try {
			return (Field[]) MH_getDeclaredFieldsQuick.invokeExact(clazz, false);
		} catch (Throwable e) {
			// should never happen
			throw new RuntimeException(e);
		}
	}

	private static final int POINTER_WIDTH = UnsafeWrapper.ADDRESS_SIZE;
	private static final int OOP_WIDTH = UnsafeWrapper.ARRAY_OBJECT_INDEX_SCALE;
	private static final boolean ARE_KPS_COMPRESSED = determineIfKpsAreCompressed();
	private static final int KLASS_PTR_WIDTH;

	static {
		KLASS_PTR_WIDTH = areKpsCompressed() ? Integer.BYTES : pointerWidth();
	}

	private static final boolean ARE_OOPS_COMPRESSED = determineIfOopsAreCompressed();
	public static final int COMPRESSED_OOP_SHIFT = log2nlz(UnsafeWrapper.ARRAY_OBJECT_INDEX_SCALE) + 1;

	/**
	 * The header size of an object instance, excluding the extra 4 bytes representing an array's length.
	 */
	// pointerWidth mark word, then klass ptr
	public static final int OBJECT_HEADER_SIZE = pointerWidth() + klassPtrWidth();

	/**
	 * The width of a native pointer.
	 */
	public static int pointerWidth() {
		return POINTER_WIDTH;
	}

	/**
	 * Width of an Ordinary Object Pointer as stored in the JVM. These may be compressed (see
	 * {@link #areOopsCompressed()} on 64-bit JVMs.
	 */
	public static int oopWidth() {
		return OOP_WIDTH;
	}

	/**
	 * Width of a klass pointer.
	 */
	public static int klassPtrWidth() {
		return KLASS_PTR_WIDTH;
	}

	private static boolean determineIfOopsAreCompressed() {
		if (pointerWidth() == oopWidth()) return false;
		// this constant is specified in the Hotspot source code...
		if (Runtime
				.getRuntime()
				.maxMemory() > (2L * 1_024L * 1_024L * 1_024L)) return true;
		String osName = System.getProperty("os.name");
		// but macOS seems to follow different rules.
		if (osName != null && osName.contains("Mac OS")) return true;
		switch (CompressedOopsMode.RUNNING) {
			case ProbablyDisabled -> {
				return false;
			}
			case UnscaledNarrow -> {
				return false;
			}
			case ZeroBasedNarrow -> {
				return true;
			}
			case DisjointBaseNarrow, HeapBasedNarrow -> {
				throw new AssertionError("Unsupported OOP compression mode: " + CompressedOopsMode.RUNNING);
			}
		}
		return false;
	}

	private static final class ObjectLayoutModel {
		@SuppressWarnings({"FieldMayBeStatic", "unused"})
		private final byte b = (byte) 0;
	}

	private static boolean determineIfKpsAreCompressed() {
		// TODO: document why I didn't just use `ObjectLayoutModel.class.getDeclaredField("b")` instead.
		Field[] fs = getDeclaredFieldsQuick(ObjectLayoutModel.class);
		assert fs.length == 1 : "Unable to get fields on ObjectLayoutModel.";
		Field f = fs[0];
		assert "b".equals(f.getName()) : "Unable to find expected field on ObjectLayoutModel";
		assert f.getType() == byte.class : "Unable to find expected field on ObjectLayoutModel";
		// minus pointerWidth for mark word
		long o = UnsafeWrapper.objectFieldOffset(f) - pointerWidth();
		if (pointerWidth() == Integer.BYTES) {
			assert o == pointerWidth();
			return false;
		}
		if (o == pointerWidth()) {
			return false;
		} else if (o == Integer.BYTES) {
			return true;
		} else {
			throw new AssertionError("Unexpected field offset: " + o);
		}
	}

	public static AssertionError unexpectedOopWidth() {
		throw new AssertionError("Unexpected OOP width: " + Unfettered.oopWidth());
	}

	/**
	 * When true, the JVM is using compressed pointers (this is the default). This will only be true on 64-bit JVMs.
	 */
	public static boolean areOopsCompressed() {
		return ARE_OOPS_COMPRESSED;
	}

	public static boolean areKpsCompressed() {
		return ARE_KPS_COMPRESSED;
	}

	static {
		if (areOopsCompressed()) {
			assert oopWidth() == Integer.BYTES
					: "oopWidth is expected to be 4 bytes when CompressedOops are enabled. oopWidth=" + oopWidth();
			assert pointerWidth() == Long.BYTES
					: "CompressedOops on 32-bit? pointerWidth=" + pointerWidth() + ", oopWidth=" + oopWidth();
		}
	}

	/**
	 * Gets the OOP of the object. If CompressedOops is enabled (see {@link #areOopsCompressed()}) then the returned
	 * value will not be valid native pointer. To make it so, use {@link #uncompressOop}.
	 */
	public static long getOopOfObject(Object obj) {
		if (oopWidth() == Long.BYTES) {
			return getOopOfObjectAssume64bit(obj);
		} else if (oopWidth() == Integer.BYTES) {
			return Integer.toUnsignedLong(getOopOfObjectAssume32bit(obj));
		} else {
			throw new UnsupportedOperationException("Unsupported OOP width: " + oopWidth());
		}
	}

	/**
	 * Gets the OOP of the object and decompresses it if it is compressed.
	 */
	public static long getUncompressedOopOfObject(Object obj) {
		return uncompressOop(getOopOfObject(obj));
	}

	/**
	 * <p>Gets the OOP of the object. If CompressedOops is enabled (see {@link #areOopsCompressed()}) then the returned
	 * value will not be valid native pointer. To make it so, use {@link #uncompressOop}.</p>
	 *
	 * <p>You probably want {@link #getOopOfObjectAssume32bit(Object)} instead.</p>
	 */
	public static int getOopOfObjectAssumeCompressed(Object obj) {
		assert areOopsCompressed() : "OOPs are not compressed";
		return getOopOfObjectAssume32bit(obj);
	}

	/**
	 * Gets the OOP of the object, assuming that it is 32-bits wide.
	 */
	public static int getOopOfObjectAssume32bit(Object obj) {
		assert oopWidth() == Integer.BYTES : "OOPs are not 32-bits wide. oopWidth=" + oopWidth();
		return getOopOfObject32bit(obj);
	}

	/**
	 * Gets the OOP of the object, assuming that it is 64-bits wide.
	 */
	public static long getOopOfObjectAssume64bit(Object obj) {
		assert oopWidth() == Long.BYTES : "OOPs are not 64-bits wide. oopWidth=" + oopWidth();
		return getOopOfObject64bit(obj);
	}

	private static int getOopOfObject32bit(Object obj) {
		return UnsafeWrapper.getInt(new Object[]{obj}, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET);
	}

	private static long getOopOfObject64bit(Object obj) {
		return UnsafeWrapper.getLong(new Object[]{obj}, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET);
	}

	/**
	 * Dereferences an OOP.
	 */
	public static Object dereferenceOop(long oop) {
		if (oopWidth() == Long.BYTES) {
			return dereferenceOopAssume64bit(oop);
		} else if (oopWidth() == Integer.BYTES) {
			// TODO: document this constant because I forget where it comes from/why not using `(int) ... == ... `
			//  instead
			assert (oop & 0xffff_ffffL) == oop : "OOP is not 32-bit: 0x" + Long.toHexString(oop);
			//assert ((int) oop) == oop : "OOP is not 32-bit: 0x" + Long.toHexString(oop);
			return dereferenceOopAssume32bit((int) oop);
		} else {
			throw new UnsupportedOperationException("Unsupported OOP width: " + oopWidth());
		}
	}

	/**
	 * Dereferences an OOP, assuming that it is compressed.
	 */
	public static Object dereferenceOopAssumeCompressed(int oop) {
		assert areOopsCompressed() : "OOPs are not compressed";
		return dereferenceOopAssume32bit(oop);
	}

	/**
	 * Dereferences an OOP, assuming that it is 32-bits wide.
	 */
	public static Object dereferenceOopAssume32bit(int oop) {
		assert oopWidth() == Integer.BYTES : "OOPs are not 32-bits wide. oopWidth=" + oopWidth();
		int[] slot = (int[]) UnsafeWrapper.allocateUninitializedArray(int.class, 1);
		UnsafeWrapper.putInt(slot, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET, oop);
		return UnsafeWrapper.getReference(slot, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET);
	}

	/**
	 * Dereferences an OOP, assuming that it is 64-bits wide.
	 */
	public static Object dereferenceOopAssume64bit(long oop) {
		assert oopWidth() == Long.BYTES : "OOPs are not 64-bits wide. oopWidth=" + oopWidth();
		long[] slot = (long[]) UnsafeWrapper.allocateUninitializedArray(long.class, 1);
		UnsafeWrapper.putLong(slot, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET, oop);
		return UnsafeWrapper.getReference(slot, UnsafeWrapper.ARRAY_OBJECT_BASE_OFFSET);
	}

	/**
	 * If CompressesOops are enabled (see {@link #areOopsCompressed()}), returns the {@code oop} with compression
	 * removed. Otherwise, returns the {@code oop} itself.
	 */
	public static long uncompressOop(long oop) {
		if (areOopsCompressed()) {
			return oop << COMPRESSED_OOP_SHIFT;
		} else {
			return oop;
		}
	}

	/**
	 * If CompressesOops are enabled (see {@link #areOopsCompressed()}), returns the {@code oop} with compression
	 * applied. Otherwise, returns the {@code oop} itself.
	 */
	public static long compressOop(long oop) {
		if (areOopsCompressed()) {
			return oop >>> COMPRESSED_OOP_SHIFT;
		} else {
			return oop;
		}
	}

	/**
	 * If {@code enableCompression}, returns the {@code ptr} with compression removed. Otherwise, returns the
	 * {@code ptr} itself.
	 */
	public static long maybeUncompressPtr(long oop, boolean enableCompression) {
		if (enableCompression) {
			return oop << COMPRESSED_OOP_SHIFT;
		} else {
			return oop;
		}
	}

	/**
	 * If {@code enableCompression}, returns the {@code ptr} with compression applied. Otherwise, returns the
	 * {@code ptr} itself.
	 */
	public static long maybeCompressPtr(long oop, boolean enableCompression) {
		if (enableCompression) {
			return oop >>> COMPRESSED_OOP_SHIFT;
		} else {
			return oop;
		}
	}

	/**
	 * Returns the maximum offset of the class's fields. This is not an efficient operation. Callers are strongly
	 * encouraged to calculate the result once and store it for reuse.
	 */
	public static long classMaxFieldOffset(Class<?> clazz) {
		if (clazz.isPrimitive()) throw new IllegalArgumentException("Cannot calculate max field offset of a " +
				"primitive");
		if (clazz.isArray()) throw new IllegalArgumentException("Cannot calculate max field offset of an array");
		long maxOffset = -1L;
		do {
			for (var field : getDeclaredFieldsQuick(clazz)) {
				maxOffset = Math.max(UnsafeWrapper.objectFieldOffset(field), maxOffset);
			}
			clazz = clazz.getSuperclass();
		} while (clazz != null);
		return maxOffset;
	}

	/**
	 * Returns the total length of all the class's fields. This is not an efficient operation. Callers are strongly
	 * encouraged to calculate the result once and store it for reuse.
	 */
	public static long classFieldsLength(Class<?> clazz) {
		if (clazz.isPrimitive()) throw new IllegalArgumentException("Cannot calculate max field offset of a " +
				"primitive");
		if (clazz.isArray()) throw new IllegalArgumentException("Cannot calculate max field offset of an array");

		long maxOffset = -1;
		Field maxField = null;

		do {
			for (var field : getDeclaredFieldsQuick(clazz)) {
				if (Modifier.isStatic(field.getModifiers())) continue;
				long offset = UnsafeWrapper.objectFieldOffset(field);
				assert offset != maxOffset : "Illegal duplicate field offset";
				if (offset > maxOffset) {
					maxOffset = offset;
					maxField = field;
				}
			}
			clazz = clazz.getSuperclass();
		} while (clazz != null);

		if (maxField == null) return 0;

		int fieldLen;
		var fieldTy = maxField.getType();
		if (fieldTy.isPrimitive()) {
			if (fieldTy == byte.class || fieldTy == boolean.class) fieldLen = 1;
			else if (fieldTy == short.class || fieldTy == char.class) fieldLen = 2;
			else if (fieldTy == int.class || fieldTy == float.class) fieldLen = 4;
			else if (fieldTy == long.class || fieldTy == double.class) fieldLen = 8;
			else throw new AssertionError("Illegal field type: " + fieldTy);
		} else {
			fieldLen = oopWidth();
		}
		return maxOffset + fieldLen;
	}

	public static long getKlassPointer(long oop) {
		// add pointerWidth for mark word
		if (klassPtrWidth() == Integer.BYTES) {
			return UnsafeWrapper.getInt(oop + pointerWidth());
		} else {
			assert klassPtrWidth() == Long.BYTES : "Unexpected klass ptr width: " + klassPtrWidth();
			return UnsafeWrapper.getLong(oop + pointerWidth());
		}
	}
}
